// This code was compiled using nvcc:
// nvcc -ptx p_to_T_dist_and_bary.cu -arch=sm_52 -ccbin="C:\Program Files (x86)\Microsoft Visual Studio 12.0\VC\bin\cl.exe"

#include "cuda.h"

// This code is used to project points on each triangle in a set of triangles
// normE0 - square norm of every first edge in a triangle (vertex #2 - vertex #1)
// dotE0E1 - dot product of every first (vertex #2 - vertex #1) and second (vertex #3 - vertex #1) edges in a triangle (length is the number of triangles)
// normE1 - square norm of every second edge in a triangle (vertex #3 - vertex #1)
// dotE0D - matrix of size #triangles x #points
//          for every triangle with first vertex P1 and first edge E0, and for every point, E0 * (P1 - point)
// dotE1D - matrix of size #triangles x #points
//          for every triangle with first vertex P1 and second edge E1, and for every point, E1 * (P1 - point)
// normD - matrix of size #triangles x #points, || point - P1 ||^2
// P1 - #triangles x dim, first vertex of each triangle
// bary - output, barycentric coordinates of the projection of each point on each triangle
// dim - points dimension (if in 3D, 3, if embedding space usually 8)
// dist - the distance between each point and triangle
void __global__ dist_Point_to_Triangle(double const * const normE0, 
    double const * const dotE0E1, double const * const normE1, 
    double const * const dotE0D, double const * const dotE1D, 
    double const * const normD, double const * const P1, int nf, int np, 
    double *bary, int dim, double *dist)
{
	int index = blockDim.x * blockIdx.x + threadIdx.x;
	int fi = index / np, pi = index % np, pfi = pi * nf + fi;
	if (fi >= nf || pi >=np) return;
	
	double a = normE0[fi], b = dotE0E1[fi], c = normE1[fi], d = dotE0D[pfi], e = dotE1D[pfi], f = normD[pfi];

	if (a*c > b*b - 1e-8 && a*c < b*b + 1e-8 ||
		a*c > -b*b - 1e-8 && a*c < -b*b + 1e-8)
	{
		dist[pfi] = 1e10; // ignore zero area triangles

		bary[pfi] = 1;
		bary[np*nf + pfi] = 0;
		bary[2 * np*nf + pfi] = 0;
	}

	double det = a*c - b*b, s = b*e - c*d, t = b*d - a*e;
	double sqrDistance = -1;

	if (s + t <= det)	{
		if (s < 0)  {
			if (t < 0) {
				//region4
				if (d < 0) {
					if (-d >= a) {
						s = 1;
						sqrDistance = a + 2 * d + f;
					}
					else {
						s = -d / a;
						sqrDistance = d*s + f;
					}
				}
				else {
					s = 0;
					if (e >= 0) {
						t = 0;
						sqrDistance = f;
					}
					else {
						if (-e >= c) {
							t = 1;
							sqrDistance = c + 2 * e + f;
						}
						else {
							t = -e / c;
							sqrDistance = e*t + f;
						}
					}
				} // of region 4
			}
			else {
				// region 3
				s = 0;
				if (e >= 0) {
					t = 0;
					sqrDistance = f;
				}
				else {
					if (-e >= c) {
						t = 1;
						sqrDistance = c + 2 * e + f;
					}
					else {
						t = -e / c;
						sqrDistance = e*t + f;
					}
				}
			} // of region 3
		}
		else {
			if (t < 0) {
				// region 5
				t = 0;
				if (d >= 0) {
					s = 0;
					sqrDistance = f;
				}
				else {
					if (-d >= a) {
						s = 1;
						sqrDistance = a + 2 * d + f;
					}
					else {
						s = -d / a;
						sqrDistance = d*s + f;
					}
				}
			}
			else
			{
				// region 0
				double invDet = 1 / det;
				s = s*invDet;
				t = t*invDet;
				sqrDistance = s*(a*s + b*t + 2 * d) + t*(b*s + c*t + 2 * e) + f;
			}
		}
	}
	else {
		if (s < 0) {
			// region 2
			double tmp0 = b + d, tmp1 = c + e;
			if (tmp1 > tmp0) // minimum on edge s+t=1
			{
				double numer = tmp1 - tmp0, denom = a - 2 * b + c;
				if (numer >= denom) {
					s = 1;
					t = 0;
					sqrDistance = a + 2 * d + f; 
				}
				else {
					s = numer / denom;
					t = 1 - s;
					sqrDistance = s*(a*s + b*t + 2 * d) + t*(b*s + c*t + 2 * e) + f;
				}
			}
			else          // minimum on edge s=0
			{
				s = 0;
				if (tmp1 <= 0) {
					t = 1;
					sqrDistance = c + 2 * e + f;
				}
				else {
					if (e >= 0) {
						t = 0;
						sqrDistance = f;
					}
					else {
						t = -e / c;
						sqrDistance = e*t + f;
					}
				}
			} // of region 2
		}
		else {
			if (t < 0) {
				//region6
				double tmp0 = b + e, tmp1 = a + d;
				if (tmp1 > tmp0)
				{
					double numer = tmp1 - tmp0, denom = a - 2 * b + c;
					if (numer >= denom) {
						t = 1;
						s = 0;
						sqrDistance = c + 2 * e + f;
					}
					else {
						t = numer / denom;
						s = 1 - t;
						sqrDistance = s*(a*s + b*t + 2 * d) + t*(b*s + c*t + 2 * e) + f;
					}
				}
				else  {
					t = 0;
					if (tmp1 <= 0) {
						s = 1;
						sqrDistance = a + 2 * d + f;
					}
					else {
						if (d >= 0) {
							s = 0;
							sqrDistance = f;
						}
						else {
							s = -d / a;
							sqrDistance = d*s + f;
						}
					}
				}
				//end region 6
			}
			else {
				// region 1
				double numer = c + e - b - d;
				if (numer <= 0) {
					s = 0;
					t = 1;
					sqrDistance = c + 2 * e + f;
				}
				else {
					double denom = a - 2 * b + c;
					if (numer >= denom) {
						s = 1;
						t = 0;
						sqrDistance = a + 2 * d + f;
					}
					else {
						s = numer / denom;
						t = 1 - s;
						sqrDistance = s*(a*s + b*t + 2 * d) + t*(b*s + c*t + 2 * e) + f;
					}
				} // of region 1
			}
		}
	}

	// account for numerical round-off error
	if (sqrDistance < 0) {
		sqrDistance = 0;
	}



	dist[pfi] = sqrt(sqrDistance);

	bary[pfi] = (1 - s - t);
	bary[np*nf + pfi] = s;
	bary[2*np*nf + pfi] = t;
}

