// compiled using mexcuda

#include "mex.h"
#include "gpu/mxGPUArray.h"

// convenient access to a GPU array arr, with row_num rows, row i, column j
double __device__ get_entry(double const * const arr, int row_num, int i, int j)
{
	return arr[row_num * j + i];
}

// Input:
// vertices - vertex coordinates (nv x dim)
// faces - 3 vertex indices per face
// nf - number of faces
// nv - number of vertices
// dim - dimension of vertex coordinates (3 if in 3D, usually 8 for embedding space)
// Output:
// P1 - coordinates of first vertex of each face
// E0 - first edge of each face (vertex #2 - vertex #1)
// E1 - second edge of each face (vertex #3 - vertex #1)
void __global__ construct_helpers(double const * const vertices, double const * const faces, int nf,
	int nv, int dim,
	double *P1, double *E0, double *E1)
{
	int index = blockDim.x * blockIdx.x + threadIdx.x;
	int fi = index / dim;
	int j = index % dim;
	if (fi >= nf || j >= dim) return;

	int P1_i = get_entry(faces, nf, (int)fi, 0) - 1, P2_i = get_entry(faces, nf, (int)fi, 1) - 1, P3_i = get_entry(faces, nf, (int)fi, 2) - 1;
	if (P1_i >= nv || P2_i >= nv || P3_i >= nv)
		return;

	P1[j*nf + fi] = get_entry(vertices, nv, P1_i, j);

	E0[j*nf + fi] = get_entry(vertices, nv, P2_i, j) - P1[j*nf + fi];
	E1[j*nf + fi] = get_entry(vertices, nv, P3_i, j) - P1[j*nf + fi];
}

// input: vertices, faces (GPU arrays). output: P1, E0, E1
void mexFunction(int nlhs, mxArray *plhs[],
	int nrhs, mxArray const *prhs[])
{
	char const * const errId = "parallel:gpu:mexGPUExample:InvalidInput";
	char const * const errMsg = "Invalid input to MEX file.";

	int const threadsPerBlock = 1024;
	int blocksPerGrid;

	/* Initialize the MathWorks GPU API. */
	mxInitGPU();

	/* Throw an error if the input is not a GPU array. */
	if ((nrhs != 2) || !(mxIsGPUArray(prhs[0])) || !(mxIsGPUArray(prhs[1]))) {
		mexErrMsgIdAndTxt(errId, errMsg);
	}

	mxGPUArray const *vertices;
	mxGPUArray const *faces;
	double const *d_vertices;
	double const *d_faces;
	double *d_P1;
	double *d_E0;
	double *d_E1;
	mxGPUArray *P1, *E0, *E1;
	mwSize nv, nf, dim;

	vertices = mxGPUCreateFromMxArray(prhs[0]);
	nv = mxGPUGetDimensions(vertices)[0];
	dim = mxGPUGetDimensions(vertices)[1];
	faces = mxGPUCreateFromMxArray(prhs[1]);
	nf = mxGPUGetDimensions(faces)[0];

	if (mxGPUGetClassID(vertices) != mxDOUBLE_CLASS) {
		mexErrMsgIdAndTxt(errId, errMsg);
	}
	if (mxGPUGetClassID(faces) != mxDOUBLE_CLASS) {
		mexErrMsgIdAndTxt(errId, errMsg);
	}
	

	d_vertices = (double const *)(mxGPUGetDataReadOnly(vertices));
	d_faces = (double const *)(mxGPUGetDataReadOnly(faces));

	mwSize dims[2] = { nf, dim };
	P1 = mxGPUCreateGPUArray(2,
		dims,
		mxGPUGetClassID(faces),
		mxGPUGetComplexity(faces),
		MX_GPU_INITIALIZE_VALUES);

	E0 = mxGPUCreateGPUArray(2,
		dims,
		mxGPUGetClassID(faces),
		mxGPUGetComplexity(faces),
		MX_GPU_INITIALIZE_VALUES);

	E1 = mxGPUCreateGPUArray(2,
		dims,
		mxGPUGetClassID(faces),
		mxGPUGetComplexity(faces),
		MX_GPU_INITIALIZE_VALUES);
	
	d_P1 = (double *)(mxGPUGetData(P1));
	d_E0 = (double *)(mxGPUGetData(E0));
	d_E1 = (double *)(mxGPUGetData(E1));


	blocksPerGrid = ((nf*dim) + threadsPerBlock - 1) / threadsPerBlock;
	construct_helpers << <blocksPerGrid, threadsPerBlock >> >(d_vertices, d_faces, nf, nv, dim,
		d_P1, d_E0, d_E1);

	plhs[0] = mxGPUCreateMxArrayOnGPU(P1);
	plhs[1] = mxGPUCreateMxArrayOnGPU(E0);
	plhs[2] = mxGPUCreateMxArrayOnGPU(E1);

	mxGPUDestroyGPUArray(vertices);
	mxGPUDestroyGPUArray(faces);
}
